﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;

public class JudgeManager : MonoBehaviour
{
    private static JudgeManager instance;
    public static JudgeManager Instance
    {
        get
        {
            if (instance == null)
            {
                instance = FindObjectOfType(typeof(JudgeManager))
                    as JudgeManager;
            }

            return instance;
        }
    }
    
    float latency = 225f;
    float scrollMultiplier = 1.0f;
    float startOffset = -5f;

    bool IsGameEnd;

    private GameObject offset;
    private GameObject noteobj, smobj, lmobj;
    private MotionGageManager motionGageManager;
    private GameObject judgeText;

    private GameObject motionSampleDisplayPrefab;

    private Vector3 initialPos;

    private float MsPerBeat
    {
        get
        {
            return 60 * 1000f / GameManager.Instance.CurrentTrack.BPM;
        }
    }
    
    private float ScrollSpeed
    {
        get
        {
            return scrollMultiplier * MsPerBeat / 1000f;
        }
    }

    void SetObjectRef()
    {
        offset = GameObject.Find("Offset");
        noteobj = GameObject.Find("Noteobj");
        smobj = GameObject.Find("SMobj");
        lmobj = GameObject.Find("LMobj");
        judgeText = GameObject.Find("Judge");
        motionGageManager = new MotionGageManager
        { motionGuage = GameObject.Find("Motion Guage") };
    }

    void JudgeInit()
    {
        initialPos = offset.transform.position;
        judgeText.SetActive(false);
        motionGageManager.ResetGuage();
        IsGameEnd = false;
    }

    void LoadGameObjects()
    {
        motionSampleDisplayPrefab = Resources.Load("Motion Sample Display") as GameObject;
        GameManager.Instance.CurrentTrack.Notes.ForEach(LoadNote);
        Instantiate(GameManager.Instance.defaultSound);
    }
    
    // Use this for initialization
    void Start()
    {
        SetObjectRef();
        JudgeInit();
        LoadGameObjects();
    }

    // Update is called once per frame
    void Update()
    {
        float timing = motionGageManager.UpdateTime(); ;

        offset.transform.position = Vector3.left * timing * ScrollSpeed;
        
        motionGageManager.UpdateGuage();

        if (noteobj.transform.childCount == 0 ||
            IsGameEnd)
            Invoke("ShowResult", 2f);

        JudgeNote(timing);
    }

    void JudgeNote(float timing)
    {
        Note note = GetLastNote(timing);
        if (note == null)
            return;

        if (note.IsLong)
            JudgeLongNote(note, timing);
        else
            JudgeShortNote(note, timing);
    }

    Note GetLastNote(float timing)
    {
        while (noteobj.transform.childCount != 0)
        {
            var note = noteobj.transform.GetChild(0).gameObject
                              .GetComponent<Note.Controller>().Instance;
            if (!Judge.IsPastNote(note, timing))
                return note;
            SetJudge(Judge.MISS);
            DeactivateNote(note);
        }

        return null;
    }

    void JudgeShortNote(Note note, float timing)
    {
        if (NoteCondition.IsWrongInput(note))
            WrongNoteProc(note);
        else if (NoteCondition.IsShortNoteEntered(note))
            EnteredNoteProc(note, timing);
    }

    void JudgeLongNote(Note note, float timing)
    {
        if (NoteCondition.IsWrongInput(note))
            WrongNoteProc(note);
        else if (NoteCondition.IsLongNoteStartCorrectly(note))
            EnteredNoteProc(note, timing);
        else if (NoteCondition.IsLongNoteHoldCorrectly(note))
            ContinuingNoteProc(note, timing);
        else if (NoteCondition.IsLongNoteFinishCorrectly(note, timing))
            CorrectlyStoppedNoteProc(note, timing);
        else if (NoteCondition.IsLongNoteFinishIncorrectly(note, timing))
            IncorrectlyStoppedNoteProc(note, timing);
    }

    void WrongNoteProc(Note note)
    {
        SetJudge(Judge.MISS);
        DeactivateNote(note);

        if (!(note is MotionNote))
        {
            IsGameEnd = true;
            Invoke("ShowResult", 2f);
        }
    }

    void EnteredNoteProc(Note note, float timing)
    {
        SetJudge(Judge.TestJudge(note, timing));

        if (!note.IsLong)
            DeactivateNote(note);
    }

    void ContinuingNoteProc(Note note, float timing)
    {
        bool isNoteEnd = Judge.IsNoteEnd(note, timing);
        var judge = Judge.TestJudge(note, timing, isNoteEnd);
        SetJudge(judge);

        if (isNoteEnd ||
            NoteCondition.IsNoteBroken(note, timing, judge))
            DeactivateNote(note);
    }

    void CorrectlyStoppedNoteProc(Note note, float timing)
    {
        SetJudge(Judge.TestJudge(note, timing, true));
        DeactivateNote(note);
    }

    void IncorrectlyStoppedNoteProc(Note note, float timing)
    {
        var judge = Judge.TestJudge(note, timing, note.Activated);
        SetJudge(Judge.IsNoteComboBroken(judge) ? judge : Judge.MISS);
        DeactivateNote(note);
    }

    private void DeactivateNote(Note note)
    {
        note.Activated = false;
        note.Component.transform.SetParent(offset.transform);
        note.Component.Deactivate();
    }

    private void SetJudge(Judge judge, bool isMotion = false)
    {
        if (!judge.IsBreak)
            ++GameManager.Instance.Combo;
        else
            GameManager.Instance.Combo = 0;
        GameManager.Instance.Score += judge.Score;
        GameManager.Instance.JudgeCount[judge]++;

        Debug.Log(judge.Name + (isMotion ? " Motion" : " Note") + " Combo: " + GameManager.Instance.Combo);

        judgeText.SetActive(true);
        judgeText.GetComponent<Text>().text = judge.Name;
        judgeText.GetComponent<Text>().color = judge.Color;
    }

    void LoadNote(Note note)
    {
        GameObject obj = Instantiate(
                Resources.Load(note.Type.ToString(), typeof(GameObject)),
                noteobj.transform)
                as GameObject;

        obj.AddComponent<Note.Controller>().Instance = note;

        if (note.Type == NoteType.MeasureLine || note.Type == NoteType.BeatLine)
            obj.transform.SetParent(offset.transform);

        if (note.IsLong)
            StretchLongNote(note, obj);

        obj.transform.position += initialPos +
            (Vector3.right * (note.StartTiming * ScrollSpeed));
    }

    void StretchLongNote(Note note, GameObject obj)
    {
        float length = note.Length * ScrollSpeed;

        var holdTransform = obj.transform.GetChild(1)
            .GetComponent<RectTransform>();
        var endTransform = obj.transform.GetChild(2)
            .GetComponent<RectTransform>();

        holdTransform.SetSizeWithCurrentAnchors(
            RectTransform.Axis.Horizontal, length);
        holdTransform.position += (Vector3.right * length / 2);
        endTransform.position += (Vector3.right * length);
    }

    private bool onResult = false;

    private void ShowResult()
    {
        if (!onResult)
        {
            GameManager.Instance.SceneTransition("Result", true);
            onResult = true;
        }
    }

    private class NoteCondition
    {
        public static Dictionary<string, InputStatus> WrongInput(Note note)
        {
            if (note is MotionNote)
                return new Dictionary<string, InputStatus>
                {
                    { "short", InputManager.Instance.ShortMotionStat },
                    { "long",  InputManager.Instance.LongMotionStat  }
                };
            return new Dictionary<string, InputStatus>
            {
                { "short", InputManager.Instance.ShortButtonStat },
                { "long",  InputManager.Instance.LongButtonStat  }
            };
        }

        public static Dictionary<string, InputStatus> JudgeInput(Note note)
        {
            if (note is MotionNote)
                return new Dictionary<string, InputStatus>
                {
                    { "short", InputManager.Instance.ShortMotionToInput((note as MotionNote).MotionName) },
                    { "long",  InputManager.Instance.LongMotionToInput((note as MotionNote).MotionName)  }
                };
            return new Dictionary<string, InputStatus>
            {
                { "short", InputManager.Instance.ShortButtonStat },
                { "long",  InputManager.Instance.LongButtonStat  }
            };
        }

        public static bool IsShortNoteEntered(Note note)
        {
            return JudgeInput(note)["short"] == InputStatus.Entered;
        }

        public static bool IsLongNoteStartCorrectly(Note note)
        {
            return JudgeInput(note)["long"] == InputStatus.Entered &&
                   !note.Activated;
        }

        public static bool IsLongNoteHoldCorrectly(Note note)
        {
            var stat = JudgeInput(note);
            return (
                        stat["long"] == InputStatus.Entered ||
                        stat["long"] == InputStatus.Continuing
                   ) &&
                   note.Activated;
        }

        public static bool IsLongNoteFinishCorrectly(Note note, float timing)
        {
            return JudgeInput(note)["long"] == InputStatus.Stopped &&
                   note.Activated &&
                   Judge.IsNoteEnd(note, timing);
        }

        public static bool IsLongNoteFinishIncorrectly(Note note, float timing)
        {
            var stat = JudgeInput(note);
            return stat["long"] == InputStatus.Stopped &&
                   (
                        !note.Activated ||
                        !Judge.IsNoteEnd(note, timing)
                   );
        }

        public static bool IsWrongInput(Note note)
        {
            var stat = WrongInput(note);
            return stat["short"] == InputStatus.Entered ||
                   stat["long"] == InputStatus.Entered ||
                   stat["long"] == InputStatus.Continuing;
        }

        public static bool IsNoteBroken(Note note, float timing, Judge judge)
        {
            return note.IsLong &&
                   !Judge.IsNoteEnd(note, timing) &&
                   (judge == Judge.BAD || judge == Judge.MISS);
        }
    }
}

class MotionGageManager : MonoBehaviour
{
    public GameObject motionGuage;
    private float elapsedMotion;
    private float motionTimeout;
    private float elapsedTime = 0;

    public void ResetGuage(float timeout = 0f)
    {
        if (timeout <= 0f)
        {
            elapsedMotion = 0f;
            motionGuage.transform.parent.gameObject.SetActive(false);
            return;
        }

        motionTimeout = timeout;
        motionGuage.transform.parent.gameObject.SetActive(true);
        motionGuage.SetActive(true);
    }
    public void UpdateGuage()
    {
        if (motionGuage.transform.parent.gameObject.activeInHierarchy)
            UpdateTime();

        if (elapsedMotion >= motionTimeout)
            ResetGuage();
        else
            motionGuage.GetComponent<Image>().fillAmount = elapsedMotion / motionTimeout;
    }

    public float UpdateTime()
    {
        elapsedTime += Time.deltaTime * 1000;

        return elapsedTime;
    }
}